function [Rguard,Rguard_noInt,split,RbeforeInt] = guardProjection(obj,hSpace,A,B,loc,partialIntersectionFlag,options)
% hyperplaneMap - computes the reachable set of the system within a 
% location, detects the guard set that is hit and computes the new set on
% the hyperplane and the subsequent mapping.
%
% Syntax:  
%    [TP,R,activeGuards,Rjump,Rcont] =
%    hyperplaneMap(obj,tStart,R0,options)
%
% Inputs:
%    obj - location object
%    tStart - start time
%    R0 - initial reachable set
%    options - options struct
%
% Outputs:
%    TP - time point struct; e.g. contains time vector of minimum times for reaching guard sets
%    R - cell array of reachable sets
%    activeGuards - active guards
%    Rjump - reachable set after jump according to the reset map
%    Rcont - reachable set due to continuous evolution without guard or
%    invariant inmtersection
%
% Example: 
%
% Other m-files required: initReach, reach, potOut, potInt, guardIntersect,
% reset
% Subfunctions: none
% MAT-files required: none
%
% See also: none

% Author:       Matthias Althoff
% Written:      23-August-2013
% Last update:  26-August-2013
%               28-August-2013
%               06-September-2013
%               10-September-2013
%               11-September-2013
% Last revision:---

%------------- BEGIN CODE --------------


% %separating halspace
% c_new = - (hSpace.c.'*A).';
% d_new = hSpace.c.'*B*options.u;
% h_new = halfspace(c_new, d_new);

%TEMPORARY: convert constrained halfspaces to halfspaces
if isa(hSpace,'constrainedHyperplane')
    hSpace = hSpace.h;
end


%init
Rguard_noInt = [];
split = 0;
RbeforeInt = [];

%partial intersection?
if ~partialIntersectionFlag
    
    %determine switching time interval
    [tmin,tmax,t_hit,t_total,RbeforeInt] = switchingTimeInterval(obj,loc,hSpace,options);
    
    %if there is an intersection
    if ~isempty(RbeforeInt)

        %get data
        x0 = center(RbeforeInt);
        u = B*options.u;

        %compute f
        dim = length(A);
        I = eye(dim);
        %Theta
        Theta = A^2*t_hit/2;
        for i=3:options.taylorTerms
            Theta = Theta + A^i*t_hit^(i-1)/factorial(i);
        end
        %Gamma
        Gamma = I;
        for i=1:(options.taylorTerms-1)
            Gamma = Gamma + A^i*t_hit^i/factorial(i+1);
        end
        f = Theta*x0 + Gamma*u;

        %use reduced zonotope 
        Rred = reduce(RbeforeInt,'girard',options.errorOrder);

        % flow manipulation
        %[f, f_cell] = flowManipulation_works(A,RbeforeInt,x0,f,hSpace);
        %[f, f_cell] = flowManipulation_perpendicular(A,Rred,x0,f,hSpace);
        [f, f_cell] = flowManipulation_inDir(A,Rred,x0,f,hSpace);
        flow = f_cell{1};

        %compute terms of Taylor series for non-error part
        [k2,L2,Q2,lambdaQuad_inv_zono,Lambda_int] = mapsQuad(hSpace, A, f, x0, Rred);

        %compute trajectory error
        [c_tilde, d_tilde, E_err] = trajectoryError(A,f_cell,u,RbeforeInt,tmin,tmax,t_hit,options.taylorTerms,hSpace.c);

        %compute terms of Taylor series for error part
        c_tilde_mid = mid(c_tilde);
        d_tilde_mid = mid(d_tilde);
        c_tilde_rad = rad(c_tilde);
        d_tilde_rad = rad(d_tilde);
        [k2_mid,L2_mid,Q2_mid,lambdaQuad_inv_zono_mid,Lambda_int_mid] = mapsQuad(hSpace, A, f, x0, Rred, c_tilde_mid, d_tilde_mid); 
        [k2_rad,L2_rad,Q2_rad,lambdaQuad_inv_zono_rad,Lambda_int_rad] = mapsQuad(hSpace, A, f, x0, Rred, c_tilde_rad, d_tilde_rad); % <-- needs to be corrected; Rred has to be bounded by a box
        k2_err = zonotope([k2_mid,abs(k2_rad)]);
        L2_err = midrad(L2_mid,abs(L2_rad));
        for iDim = 1:length(Q2_mid)
            Q2_err{iDim} = concatenate(Q2_mid{iDim},Q2_rad{iDim});
        end


        %individual reachable sets 
        dR = RbeforeInt + (-x0);
        dRred = Rred + (-x0);

        %linear part
        Rlin = (I+L2)*dR;
        Rlin_err = L2_err*dR;


        %quadratic part
        Rquad_zono_aux = quadraticMultiplication_zono(dRred,Q2);
        Rquad_zono = 1/2*lambdaQuad_inv_zono*Rquad_zono_aux;

        Rquad_zono_aux_err = quadraticMultiplication_zono(dRred,Q2_err);
        Rquad_zono_err = 1/2*lambdaQuad_inv_zono*Rquad_zono_aux_err;


        %combine results
        Rhit_const = x0 + k2 + exactPlus(Rlin, Rquad_zono);
        Rhit_err = k2_err + Rlin_err + Rquad_zono_err + E_err;


    %     %for test
    %     Zk = k2 + zonotope([zeros(dim,1), 0.01*eye(dim)]);

        %compute mapping error
        %[delta_Rhit,E] = mappingError(A,f_cell,u,RbeforeInt,tmin,tmax,t_hit,options.taylorTerms,Lambda_int,hSpace.c);
        %delta_Rhit2 = mappingError_split(A,f,u,RbeforeInt,tmin,tmax,t_hit,options.taylorTerms,Lambda_int,hSpace.c,0.1*options.maxTimeStep);


        %compute set on hyperplane
        %Rtmp = Rhit_const + delta_Rhit;
        Rtmp = Rhit_const + Rhit_err;

        %reduce result
        Rtmp = reduce(Rtmp,'girard',options.zonotopeOrder);

        %project result
        Rtmp = project(hSpace, Rtmp);
        
        %save result
        Rguard{1} = Rtmp;


    %     intervalhull(Rquad_zono)
    %     intervalhull(delta_Rhit)
    %     intervalhull(Rhit_err)
        
        %projected errors
        Rhit_err_proj = project(hSpace, Rhit_err);
        Rquad_zono_proj = project(hSpace, Rquad_zono);

        %error index
        eLen = edgeLength(intervalhull(Rhit_err_proj) + intervalhull(Rquad_zono_proj));
        errorInd = max(eLen(1:2));
        %errorInd = max(edgeLength(intervalhull(Rhit_err)));
        
        if errorInd > options.maxProjectionError
        %if errorInd > 0.1
        %if errorInd > 0.05
            split = 1
        end

    %     %test of constant flow
    %     figure
    %     Rred = reduce(Rhit_const,'girard',10);
    %     plot(Rred,[1 2],'g');
    %     plot(RbeforeInt,[1 2],'g');
    %     for i = 1:100
    %         if i < 50
    %             x0 = randPointExtreme(RbeforeInt);
    %         else
    %             x0 = randPoint(RbeforeInt);
    %         end
    %         flowVec = A*x0 + f;
    %         xNew = x0 + tmax*flowVec;
    %         plot([x0(1),xNew(1)],[x0(2),xNew(2)],'r');
    %     end


    %     %test of exact flow
    %     figure
    %     Rred = reduce(Rguard,'girard',10);
    %     plot(Rred,[1 2],'g');
    %     plot(RbeforeInt,[1 2],'g');
    %     
    %     %get contDynamics
    %     contDynamics = get(loc,'contDynamics');
    %     
    %     %obtain data from location
    %     inv = intervalhull(1e6*ones(length(options.x0),1)*[-1 1]);
    %     tran_old = get(loc, 'transition');
    %     reset = get(tran_old{1}, 'reset');
    %     tran{1}=transition(hSpace,reset,0,'a','b'); 
    %     %create dummy location
    %     loc = location('dummyLoc',inv,tran,contDynamics);  
    %     eventOptions = odeset('Events',eventFcn(loc));
    %     opt = odeset(eventOptions);
    %     
    %     for i = 1:100
    %         if i < 50
    %             options.x0 = randPointExtreme(RbeforeInt);
    %         else
    %             options.x0 = randPoint(RbeforeInt);
    %         end
    %         
    %         %simulate continuous dynamics
    %         [contDynamics,t,x] = simulate(contDynamics,options,tmin,tmax,options.x0,opt);
    %         
    %         plot(x(:,1),x(:,2),'r');
    %     end


    %     %test delta_Rhit
    %     figure
    %     hold on
    %     
    %     
    %     %get contDynamics
    %     contDynamics = get(loc,'contDynamics');
    %     
    %     %simulate
    %     stepsizeOptions = odeset('MaxStep',0.05*(options.tStart-options.tFinal));
    %     %generate overall options
    %     opt = odeset(stepsizeOptions);
    %     
    %     for i = 1:3
    %         x0 = randPointExtreme(RbeforeInt);
    %         
    %         %simulate continuous dynamics
    %         %[contDynamics,t,x] = simulate(contDynamics,options,tmin,tmax,x0,opt);
    %         [contDynamics,t,x] = simulate(contDynamics,options,0,tmax,x0,opt);
    %         
    %         %xNew = x0 + (tmax-tmin)*flowVec;
    %         xNew = x0 + (tmax)*(A*x0 + f);
    %         plot([x0(1),xNew(1)],[x0(2),xNew(2)],'g');
    %         plot(x(:,1),x(:,2),'r');
    %         
    %         Rred = reduce(E,'girard',10);
    %         Rred_err = reduce(E_err,'girard',10);
    %         %Rred2 = reduce(delta_Rhit2,'girard',10);
    %         Rtest = Rred + xNew;
    %         Rtest_err = Rred_err + xNew;
    %         %Rtest2 = Rred2 + xNew;
    %         plot(Rtest);
    %         plot(Rtest_err,[1 2],'k');
    %         
    %         %compute delta_x
    %         delta_aux = 0*x0;
    %         for i = 2:options.taylorTerms
    %             delta_aux = delta_aux + A^i/factorial(i)*(tmax^(i-1)*x0 - t_hit^(i-1)*x0);
    %         end
    %         for i = 0:options.taylorTerms
    %             delta_aux = delta_aux + A^i/factorial(i+1)*(tmax^(i)- t_hit^(i))*u;
    %         end
    %         delta_x = tmax*delta_aux;
    %         
    %         x_corr = xNew + delta_x;
    %         plot(x_corr(1), x_corr(2), 'or');
    %     end
    else
        Rguard_noInt = [];
        Rguard{1} = [];
        split = [];
        flow = [];
    end
else
    [Rguard_noInt, Rguard] = partialGuardProjection(obj,hSpace,A,B,loc,options);
end



%function which determines when the state of x'=Ax+u crosses the hypeplane
%c^T x = d with initial state x0
function t_hit = crossingTime(hSpace,x0,A,u,tc,tmin,tmax)

t_hit = tc;
t_lower = tmin;
t_upper = tmax;
t_old = inf;

while (abs(t_hit - t_old) > 1e-6*tmax)
    
    %remember time
    t_old = t_hit;
    
    %compute state solution
    x_upper = stateSolution(x0,A,u,0.5*(t_hit + t_upper));
    x_lower = stateSolution(x0,A,u,0.5*(t_hit + t_lower));

    %check if time is too large or too small
    if (abs(hSpace.c'*x_upper - hSpace.d) < abs(hSpace.c'*x_lower - hSpace.d))
        t_lower = t_hit;
        t_hit = 0.5*(t_hit + t_upper);
    else
        t_upper = t_hit;
        t_hit = 0.5*(t_hit + t_lower);
    end

end


%computes the solution for x0 under x'=Ax+u after time t
function x = stateSolution(x0,A,u,t)

%Gamma matrix for input solution using 10 Taylor terms
Gamma = eye(length(A))*t;
for i=1:10
    Gamma = Gamma + 1/factorial(i+1)*A^i*t^(i+1);
end

x = expm(A*t)*x0 + Gamma*u;

function Rcubic = cubicIntMultiplication(dR,Cint)

%obtain interval of dR
dx = interval(intervalhull(dR));

%dim
dim = length(Cint);

for i = 1:dim
    int(i,1) = infsup(0,0);
    for l = 1:dim
        for m = 1:dim
            for n = 1:dim
                int(i,1) = int(i,1) + Cint{i}(l,m,n)*dx(l)*dx(m)*dx(n);
            end
        end
    end
end

%intervalhull
IH = intervalhull([inf(int),sup(int)]);

%Rcubic
Rcubic = zonotope(IH);


function Rcubic = cubicIntMultiplication_zono(dR,C_c,C_g)

%obtain interval of dR
Zmat = get(dR,'Z');

%dim
dim = length(C_c);
gens = length(C_g(1,:));

%init Znew
Znew = zeros(dim,(gens+1)^2);

%center computation
for i = 1:dim
    for l = 1:dim
        for m = 1:dim
            for n = 1:dim
                %compute center 
                for iGen=1:length(Zmat(1,:))
                    Znew(i,iGen) = Znew(i,iGen) + C_c{i}(l,m,n)*Zmat(l,iGen)*Zmat(m,iGen)*Zmat(n,iGen);
                end
            end
        end
    end
end

%generator computation
for iMatGen = 1:gens
    for i = 1:dim
        for l = 1:dim
            for m = 1:dim
                for n = 1:dim
                    %compute center 
                    for iGen=1:length(Zmat(1,:))
                        ind = (gens+1)*iMatGen + iGen;
                        Znew(i,ind) = Znew(i,ind) + C_g{i,iMatGen}(l,m,n)*Zmat(l,iGen)*Zmat(m,iGen)*Zmat(n,iGen);
                    end
                end
            end
        end
    end
end

%Rcubic
Rcubic = zonotope(Znew);




function [f, f_cell] = flowManipulation_perpendicular(A,R0,x0,f,hSpace)

%obtain flow in center
flow = A*x0 + f;
f_cell{1} = flow;

%obtain direction vector
dirVec = flow - flow.'*hSpace.c/norm(hSpace.c)*hSpace.c/norm(hSpace.c);

%check flow direction
limitRespected = 0;

while ~limitRespected
    flowDir_IH = intervalhull(hSpace.c'*(A*R0 + f)*(1/(norm(hSpace.c)*norm(A*x0 + f))));
    
    if flowDir_IH(1,1)*flowDir_IH(1,2) < 0
        flowDir_IH_abs = 0;
    else
        [flowDir_IH_abs] = min([abs(flowDir_IH(1,1)), abs(flowDir_IH(1,2))]);
    end
    
    %flowDir_min = 1.0;
    %flowDir_min = 0.7; %<-- best when not splitting
    %flowDir_min = 0.7;
    flowDir_min = 0.2;
    %flowDir_min = 0;
    if flowDir_IH_abs < 0.8*flowDir_min

        %stretching direction
        direction = sign(flowDir_IH(1,1) + flowDir_IH(1,2)); %is center positive or negative?
        %stretching factor
        %alpha = -direction*(flowDir_min - flowDir_IH_abs)*norm(hSpace.c)*norm(A*x0 + flow)/(hSpace.c'*hSpace.c);
        alpha = -direction*0.5;

        %flow correction
        f = f + alpha*dirVec;
        flow = flow + alpha*dirVec;
    else
        limitRespected = 1;
    end
end
f_cell{2} = flow;


function [f, f_cell] = flowManipulation_inDir(A,R0,x0,f,hSpace)

%obtain flow in center
flow = A*x0 + f;
f_cell{1} = flow;

%check flow direction
limitRespected = 0;

while ~limitRespected
    flowDir_IH = intervalhull(hSpace.c'*(A*R0 + f)*(1/(norm(hSpace.c)*norm(A*x0 + f))));
    
    if flowDir_IH(1,1)*flowDir_IH(1,2) < 0
        flowDir_IH_abs = 0;
    else
        [flowDir_IH_abs] = min([abs(flowDir_IH(1,1)), abs(flowDir_IH(1,2))]);
    end
    
    %flowDir_min = 0.4;
    flowDir_min = 0.2;
    %flowDir_min = 0.1;
    %flowDir_min = 0;
    if flowDir_IH_abs < 0.8*flowDir_min

        %strecthing factor for flow correction

        direction = sign(flowDir_IH(1,1) + flowDir_IH(1,2)); %is center positive or negative?
        alpha = direction*(flowDir_min - flowDir_IH_abs)*norm(hSpace.c)*norm(A*x0 + f)/(hSpace.c'*hSpace.c);

        %flow correction
        f = f + alpha*hSpace.c;
        flow = flow + alpha*hSpace.c;
    else
        limitRespected = 1;
    end
end
f_cell{2} = flow;




function [f, f_cell] = flowManipulation_works(A,R0,x0,f,hSpace)

%obtain flow in center
flow = A*x0 + f;
f_cell{1} = flow;

%check flow direction
limitRespected = 0;

while ~limitRespected
    flowDir_IH = intervalhull(hSpace.c'*(A*R0 + flow)*(1/(norm(hSpace.c)*norm(A*x0 + flow))));
    flowDir_IH_abs = min(abs(flowDir_IH(1,1)), abs(flowDir_IH(1,2)));
    %flowDir_min = 1.0;
    %flowDir_min = 0.8;
    flowDir_min = 0.4;
    %flowDir_min = 0.2;
    %flowDir_min = 0;
    if flowDir_IH_abs < 0.8*flowDir_min

        %strecthing factor for flow correction

        direction = sign(flowDir_IH(1,1) + flowDir_IH(1,2)); %is center positive or negative?
        alpha = direction*(flowDir_min - flowDir_IH_abs)*norm(hSpace.c)*norm(A*x0 + flow)/(hSpace.c'*hSpace.c);

        %flow correction
        f = f + alpha*hSpace.c;
        flow = flow + alpha*hSpace.c;
    else
        limitRespected = 1;
    end
end
f_cell{2} = flow;






function [delta_Rhit,G] = mappingError(A,f,u,R0,tmin,tmax,tc,order,Lambda_int,n)

%compute powers of tmin and tmax
tZono = timePowers(tmin,tmax,order);

%obtain center and delta set
x0 = center(R0);
x0zono = zonotope(x0);
uzono = zonotope(u);
dR = R0 + (-x0);

%auxiliary set: new
Raux_1 = zonotope(0*x0);
Raux_2 = zonotope(0*x0);
for i=2:order
    %Raux_1 = Raux_1 + A^i/factorial(i)*(tZono{i-1}*dR + (tZono{i-1} + (-tc^(i-1)))*x0zono);
    Raux_1 = Raux_1 + A^i/factorial(i)*(tmax^(i-1)*dR + (tZono{i-1} + (-tc^(i-1)))*x0zono);
end
for i=1:order
    Raux_2 = Raux_2 + A^i/factorial(i+1)*(tZono{i} + (-tc^i))*uzono;
end

%compute remainder matrix
A_abs = abs(A);
Apower_abs = eye(length(A_abs));
M = eye(length(A_abs));

for i=1:order
    Apower_abs = Apower_abs*A_abs;
    M = M + Apower_abs*tmax^i/factorial(i);
end 
  
%determine error due to finite Taylor series
W = expm(A_abs*tmax) - M;
%compute absolute value of W for numerical stability
W = abs(W);
E = infsup(-W,W);

%final additional set
G = tZono{1}*(Raux_1 + Raux_2 + zonotope(f{1}-f{2})) + E*R0 + E*tmax*uzono;

%error due to uncertain hitting time
delta_t = interval(intervalhull((-n')*G))/Lambda_int;
delta_t_mid = mid(delta_t);
delta_t_rad{1} = rad(delta_t);
delta_t_zono = matZonotope(delta_t_mid,delta_t_rad);

%final uncertainty
delta_Rhit = G + delta_t_zono*(A*R0 + f{2});



function [c_tilde, d_tilde, E_err] = trajectoryError(A,f,u,R0,tmin,tmax,tc,order,n)

%obtain dimension
dim = length(u);

%compute powers of tmin and tmax
tZono = timePowers(tmin,tmax,order);

%obtain center and delta set
x0 = center(R0);
x0zono = zonotope(x0);
uzono = zonotope(u);
dR = R0 + (-x0);

%auxiliary matrix
A_tilde = zeros(dim);
for i=2:order
    A_tilde = A_tilde + A^i/factorial(i)*tZono{i};
end

%auxiliary vector
u_tilde = zeros(dim,1);
for i=2:order
    u_tilde = u_tilde + A^i/factorial(i)*(tZono{1} * (-tc^(i-1)))*x0zono;
end
for i=1:order
    u_tilde = u_tilde + A^i/factorial(i+1)*(tZono{i+1} + tZono{1}*(-tc^i))*uzono;
end

%compute remainder matrix
A_abs = abs(A);
Apower_abs = eye(length(A_abs));
M = eye(length(A_abs));

for i=1:order
    Apower_abs = Apower_abs*A_abs;
    M = M + Apower_abs*tmax^i/factorial(i);
end 
  
%determine error due to finite Taylor series
W = expm(A_abs*tmax) - M;
%compute absolute value of W for numerical stability
W = abs(W);
E = infsup(-W,W);

%consider E
gen{1} = rad(E);
A_tilde = A_tilde + matZonotope(mid(E), gen);
u_tilde = u_tilde + E*tmax*uzono + tZono{1}*zonotope(f{1}-f{2}); %error due to flow correction

%test
E_err = A_tilde*R0 + u_tilde;

%obtain c_tilde and d_tilde
c_tilde = A_tilde.'*n;
c_tilde = intervalMatrix(c_tilde);
c_tilde = c_tilde.int;

d_tilde = -n.'*u_tilde;
d_tilde = intervalhull(d_tilde);
d_tilde = infsup(d_tilde(1,1),d_tilde(1,2));






function tZono = timePowers(tmin,tmax,order)


%first order
tminPow(1) = tmin;
tmaxPow(1) = tmax;

tradPow(1) = 0.5*(tmaxPow(1) - tminPow(1));
tmidPow(1) = 0.5*(tmaxPow(1) + tminPow(1));

tRad{1} = tradPow(1);
tZono{1} = matZonotope(tmidPow(1),tRad);

%higher orders
for i=1:order
    tminPow(i+1) = tminPow(i)*tmin;
    tmaxPow(i+1) = tmaxPow(i)*tmax;
    
    tradPow(i+1) = 0.5*(tmaxPow(i+1) - tminPow(i+1));
    tmidPow(i+1) = 0.5*(tmaxPow(i+1) + tminPow(i+1));
    
    tRad{1} = tradPow(i+1);
    tZono{i+1} = matZonotope(tmidPow(i+1),tRad);
end


function [tmin,tmax,t_hit,t_total,RbeforeInt,x_hit] = switchingTimeInterval(obj,loc_old,h,options)

%get intersection times
%[tmin,tmax,t_total,RbeforeInt] = intersectionTime(obj,loc_old,h,options);
contDynamics = get(loc_old,'contDynamics');
[tmin,tmax,t_total,RbeforeInt] = intersectionTime_oneDirection(h,obj,contDynamics,options);

if ~isempty(RbeforeInt)

    %get contDynamics
    contDynamics = get(loc_old,'contDynamics');

    %compute t_hit
    %obtain data from location
    inv = intervalhull(1e6*ones(length(options.x0),1)*[-1 1]);
    tran_old = get(loc_old, 'transition');
    reset = get(tran_old{1}, 'reset');
    tran{1}=transition(h,reset,0,'a','b'); 
    %create dummy location
    loc = location('dummyLoc',0,inv,tran,contDynamics);  
    eventOptions = odeset('Events',eventFcn(loc));
    opt = odeset(eventOptions);
    %simulate continuous dynamics
    [contDynamics,t,x] = simulate(contDynamics,options,tmin,tmax,center(RbeforeInt),opt);

    t_hit = t(end) - tmin;
    x_hit = x(end,:).';
else
    t_hit = [];
    x_hit = [];
end




%------------- END OF CODE --------------