function [P] = polytope(varargin)
% polytope - Converts a zonotope to a polytope representation
%
% Syntax:  
%    [P] = polytope(Z)
%
% Inputs:
%    Z - zonotope object
%
% Outputs:
%    P - polytope object
%
% Example: 
%    Z=zonotope(rand(2,5));
%    P=polytope(Z);
%    plot(P);
%    hold on
%    plot(Z);
%
% Other m-files required: vertices, polytope
% Subfunctions: none
% MAT-files required: none
%
% See also: intervalhull,  vertices

% Author:       Matthias Althoff
% Written:      18-September-2007
% Last update:  26-August-2010
%               20-October-2010
% Last revision:---

%------------- BEGIN CODE --------------

if nargin==1
     Z=varargin{1};
     W{1}=eye(length(Z.Z(:,1)));
     filterLength=[];
     options.polytopeType = 'mpt';
elseif nargin==2
     Z=varargin{1};
     options=varargin{2};
     
     %transformation
     if isfield(options,'W')
        W=options.W;
     else
        W{1}=eye(length(Z.Z(:,1)));
     end
     
     %filter length
     if isfield(options,'filterLength')
        filterLength=options.filterLength;
     else
         dim=length(Z.Z(:,1));
         filterLength(1)=dim+2;
         filterLength(2)=dim+3;
     end
     
     %flag for tight conversion
     if isfield(options,'tightPolytopeConversion');
        tightConversion=options.tightPolytopeConversion;
     else
         tightConversion=1;
     end
end


%Zmapped
for i=1:length(W)
    Zmapped{i} = pinv(W{i})*Z;
end

%use simple reduction
if isempty(filterLength)
     Zred=zonotope(intervalhull(Z));
     P=parallelotope(Zred, options);
     
%filter length is specified     
else
     if tightConversion
         %solution1 (axis-aligned):
         Zred=zonotope(intervalhull(Z));
         P=parallelotope(Zred, options); 
         %solution 2 (method C):
         for i=1:length(W)
                 Zred=reduce(Zmapped{i},'methC',1,filterLength);
                 Zred = repair(Zred,Z);
                 Padd=W{i}*parallelotope(Zred, options);
                 %intersect results
                 P=P&Padd;
         end
     else
         %solution 1 (method C):
         Zred1=W{1}*reduce(Zmapped{1},'methC',1,filterLength);
         Zred1 = repair(Zred1,Z);
         vol1=volume(Zred1);
         %solution2 (axis-aligned):
         Zred2=zonotope(intervalhull(Z));
         Zred2 = repair(Zred2,Z);
         vol2=volume(Zred2);

         if vol1<vol2
            P=parallelotope(Zred1, options);
         else
            P=parallelotope(Zred2, options);
         end
     end
end

%repair zonotope if there is no length in one dimenion
function Zrep = repair(Z,Zorig)

%get length of each dimension
len = edgeLength(intervalhull(Z));

%find zero lengths
index = find(len==0);

if ~isempty(index)
    %construct zonotope to be added
    origLen = edgeLength(intervalhull(Zorig));
    origCenter = center(intervalhull(Zorig));
    
    %get Zmatrix
    Zmat = get(Z,'Z');
    
    for i=1:length(index)
        
        ind = index(i);
        %replace center
        Zmat(ind,1) = origCenter(ind);
        %replace generator value
        Zmat(ind,ind+1) = 0.5*origLen(ind);
    end

    %instantiate zonotopes
    Zrep = zonotope(Zmat);
else
    Zrep = Z;
end


%------------- END OF CODE --------------