function [P] = wrongpolytope(Z)
% polytope - Converts a zonotope to a polytope representation
%
% Syntax:  
%    [P] = polytope(Z)
%
% Inputs:
%    Z - zonotope object
%
% Outputs:
%    P - polytope object
%
% Example: 
%    Z=zonotope(rand(2,5));
%    P=polytope(Z);
%    plot(P);
%    hold on
%    plot(Z);
%
% Other m-files required: vertices, polytope
% Subfunctions: none
% MAT-files required: none
%
% See also: intervalhull,  vertices

% Author: Matthias Althoff
% Written: 14-September-2006 
% Last update: 22-March-2007
% Last revision: ---

%------------- BEGIN CODE --------------

%retrieve center of the zonotope
c=Z.Z(:,1);
%retrieve generator matrix G
G=Z.Z(:,2:end);
%get number of generators
nrOfGen=length(G(1,:));

%first two halfspaces:

%d for positive generator direction
dPos=(c+G(:,1))'*G(:,1)/norm(G(:,1));
%d for negative generator direction
dNeg=(c-G(:,1))'*G(:,1)/norm(G(:,1));
%compose d vector
d=[dPos;dNeg];
%initial C matrix:
C=[G(:,1)';-G(:,1)'];

%for each generator add two halfspaces
for i=2:nrOfGen
    %expand old generators 
    for j=1:(i-1)
        exp=G(:,i)'*G(:,j)/norm(G(:,j));
        d(2*j-1)=d(2*j-1)+abs(exp);
        d(2*j)=d(2*j)-abs(exp);
    end
    %add new halfspaces
    posVector=c;
    negVector=c;
    for j=1:(i-1)
        %determine sign of next generator
        s=sign(G(:,i)'*G(:,j));
        posVector=posVector+s*G(:,j);
        negVector=negVector-s*G(:,j);
    end    
    posVector=posVector+G(:,i);
    negVector=negVector-G(:,i);    
    dPos=posVector'*G(:,i)/norm(G(:,i));
    dNeg=negVector'*G(:,i)/norm(G(:,i));
    % update d and C
    d=[d;dPos;dNeg];
    C=[C;G(:,i)';-G(:,i)'];  
end
%get absolute values of d
d=abs(d);

P=polytope(C,d);

%------------- END OF CODE --------------